# pylint: disable=import-error,broad-except
"""Arcam Solo telnet communication library."""

import asyncio
import time
import logging
import re
import math
import uuid

from datetime import datetime
from collections.abc import Callable
from inspect import isfunction

from .commands import (
    ARCAM_COMM_END,
    ARCAM_COMM_START,
    COMMAND_CODES,
    IR_COMMAND_CODES,
    SOURCE_IR_CONTROL_MAP,
    ARCAM_QUERY_COMMANDS,
    RADIO_QUERY_COMMANDS
)
from .util import sock_set_keepalive, cancel_task, get_backoff_delay, safe_wait_for
from .parser import parse_response
from .params import CONF_ENABLED_ZONES
from ._version import __version__ as VERSION

_LOGGER = logging.getLogger(__name__)

class ArcamSolo:
    """Base Arcam Solo module."""

    def __init__(
        self,
        host,
        port,
        timeout=2,
        scan_interval=60,
        params: dict | None=None):
        """Initialise the Arcam Solo interface."""
        _LOGGER.info("Starting pyarcamsolo %s", VERSION)
        _LOGGER.debug(
            '>> ArcamSolo.__init__(host="%s", port="%s", timeout="%s", params="%s")',
            host,
            port,
            timeout,
            params
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self.scan_interval = scan_interval

        # Public props
        self.software_version = None
        self.available = False

        # Data
        self.zones = {}
        self._zone_callbacks = {}

        # Locks
        self._connect_lock = asyncio.Lock()
        self._disconnect_lock = asyncio.Lock()
        self._update_lock = asyncio.Lock()
        self._request_lock = asyncio.Lock()
        # Events
        self._update_event = asyncio.Event()
        self._response_event = asyncio.Event()
        self._response_queue = []
        self._queue_responses = False
        self._reconnect = True
        self._full_update = True
        self._last_command = None
        self._last_response = None
        self._last_updated = None
        self._reader = None
        self._writer = None
        self._listener_task = None
        self._responder_task = None
        self._reconnect_task = None
        self._updater_task = None
        self._command_queue_task = None
        self._initial_query = True
        # Stores a list of commands to run after receiving an event
        self._command_queue = []

        # Handle configuration
        self._enabled_zones = [1] # Zone 2 is optional
        if params is not None:
            self._enabled_zones = params.get(CONF_ENABLED_ZONES, [1]) # Zone 2 is optional

    def __del__(self):
        _LOGGER.debug(">> ArcamSolo.__del__()")

    def _set_socket_options(self):
        """Set socket keepalive options."""
        sock_set_keepalive(
            self._writer.get_extra_info("socket"),
            after_idle_sec=int(self._timeout),
            interval_sec=int(self._timeout),
            max_fails=3,
        )

    def set_zone_callback(
        self, zone: int, callback_id: uuid.UUID = None, callback: Callable[..., None] | None = None
    ) -> uuid.UUID | None:
        """Configure a zone callback."""
        if zone in self.zones:
            self._zone_callbacks.setdefault(zone, {})
            if callback:
                callback_id = uuid.uuid4()
                self._zone_callbacks[zone][callback_id] = callback
                return callback_id
            elif callback_id:
                if callback_id in self._zone_callbacks[zone]:
                    self._zone_callbacks[zone].pop(callback_id)
                    return None
                raise ValueError("Callback does not exist.")
            else:
                if zone in self._zone_callbacks:
                    self._zone_callbacks.pop(zone)
                    return None
                raise ValueError("Zone does not exist.")

    def _clear_zone_callbacks(self):
        """Clear any configured zone callbacks."""
        self._zone_callbacks = {}

    def _call_zone_callbacks(self, zone: int):
        """Call a configured callback."""
        if zone in self._zone_callbacks:
            callbacks = self._zone_callbacks[zone]
            for k, v in callbacks.items():
                if v:
                    v()

    @property
    def get_unique_id(self):
        """Return a unique ID."""
        return f"{self._host}:{self._port}"

    async def _responder_cancel(self):
        """Cancel any active responder task."""
        await cancel_task(self._responder_task, "responder")
        self._responder_task = None

    def _set_updated_values(self, value):
        """Set updated values from response parser."""
        _LOGGER.debug(">> ArcamSolo._set_updated_values(value=%s)")
        if value["z"] not in self.zones:
            _LOGGER.debug("Zone does not yet exist so creating one.")
            self.zones[value["z"]] = {}
        self.zones[value["z"]][value["k"]] = value["v"]
        self._call_zone_callbacks(zone=value["z"])

    async def _connection_listener(self):
        """Arcam connection listener. Parse responses and update state."""
        _LOGGER.debug(">> ArcamSolo._connection_listener() started")
        running = True
        while self.available:
            action = " listening for responses"
            try:
                response = await self._read_response()
                if response is None:
                    # Connection closed or exception
                    break
                self._last_updated = time.time()
                self._last_response = response
                if not response:
                    # Skip empty response
                    continue
                start = response.find(ARCAM_COMM_START)
                end = response.find(ARCAM_COMM_END)
                if start >= 0 and end > 0:
                    response = response[start:end+1]
                    _LOGGER.debug("received Arcam response: %s", response.hex())
                    action = " parsing response " + response.hex()
                    value = parse_response(response)
                    if isinstance(value, list):
                        for parsed in value:
                            self._set_updated_values(parsed)
                    elif isinstance(value, dict):
                        self._set_updated_values(value)
                else:
                    _LOGGER.debug("Ignoring response %s due to invalid data.", response.hex())
            except asyncio.CancelledError:
                _LOGGER.debug((">> ArcamSolo._connection_listener() cancelled"))
                running = False
                break
            except Exception as exc:
                _LOGGER.error("listener exception%s: %s",
                              action,
                              str(exc))
                # continue listening on exception

        if running and self.available:
            # Trigger disconnection
            await self.disconnect()

        _LOGGER.debug(">> ArcamSolo._connection_listener() completed")

    async def _listener_schedule(self):
        """Schedule the listener task."""
        _LOGGER.debug(">> ArcamSolo._listener_schedule()")
        await self._listener_cancel()
        self._listener_task = asyncio.create_task(self._connection_listener())

    async def _listener_cancel(self):
        """Cancel the listener task."""
        await cancel_task(self._listener_task, "listener")
        self._listener_task = None

    async def _updater_cancel(self):
        """Cancel the updater task."""
        await cancel_task(self._updater_task, "updater")
        self._updater_task = None

    async def _updater(self):
        """Perform update scan every scan_interval."""
        event = self._update_event
        while True:
            try:
                event.clear()
                await self._updater_update()
                await safe_wait_for(event.wait(), timeout=self.scan_interval)
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                break
            except Exception as exc: #pylint: disable=broad-except
                _LOGGER.error(">> ArcamSolo._updater() exception: %s", str(exc))
                break

    async def _updater_update(self):
        """Update cached status."""
        if self._update_lock.locked():
            _LOGGER.debug("Updates locked, skipping")
            return False
        if not self.available:
            _LOGGER.debug("Device not connected, skipping update.")
            return False

        _rc = True
        async with self._update_lock:
            # Update only if scan interval has passed
            now = time.time()
            full_update = self._full_update
            scan_interval = self.scan_interval
            since_updated = scan_interval + 1
            since_updated_str = "never"
            if self._last_updated:
                since_updated = now - self._last_updated
                since_updated_str = f"{since_updated:.3f}s ago"
            if full_update or not scan_interval or since_updated > scan_interval:
                _LOGGER.info(
                    "updating Arcam status (full=%s, last updated %s)",
                    full_update,
                    since_updated_str
                )
                self._last_updated = now
                self._full_update = False
                try:
                    # update time
                    await self.send_raw_command(
                        command="time",
                        data=[
                            (datetime.today().weekday()+1).to_bytes(1, 'little'),
                            datetime.now().hour.to_bytes(1, 'little'),
                            datetime.now().minute.to_bytes(1, 'little'),
                            datetime.now().second.to_bytes(1, 'little')
                        ]
                    )
                    for zone in self.zones:
                        await self._update_zone(zone)
                    await self._update_radio_data()
                except Exception as exc: #pylint: disable=broad-except
                    _LOGGER.error(
                        "could not update Arcam status: %s: %s",
                        type(exc).__name__,
                        str(exc)
                    )
                    _rc = False
            else:
                _rc = None
        if _rc is False:
            # disconnect on error
            await self.disconnect()
        return _rc

    async def _update_zone(self, zone: int):
        """Update a given zone."""
        for query in ARCAM_QUERY_COMMANDS:
            await self.send_raw_command(
                command=query,
                data=[b'\xF0'],
                zone=zone
            )

    async def _update_radio_data(self):
        """Update radio status information."""
        _LOGGER.debug(">> ArcamSolo._update_radio_data()")
        if self.source in ["AM", "FM", "DAB"]:
            for k, v in RADIO_QUERY_COMMANDS.items():
                if (k=="request_station_frequency"
                    and self.source not in ["AM", "FM"]):
                    continue
                if (k=="request_mpeg_mode" and self.source != "DAB"):
                    continue
                if (k=="request_data_rate" and self.source != "DAB"):
                    continue
                await self.send_raw_command(
                    command="radio_station_info",
                    data=[b'\xF0', v],
                    zone=1
                )


    async def _updater_schedule(self):
        """Schedule/reschedule the updater task."""
        if self.scan_interval:
            _LOGGER.debug(">> ArcamSolo._updater_schedule()")
            await self._updater_cancel()
            self._full_update = True # always perform a full update on schedule
            self._updater_task = asyncio.create_task(self._updater())

    async def connect(self, reconnect=True):
        """Open a connection to the Hi-Fi and start listener thread."""
        _LOGGER.debug(">> ArcamSolo.connect() started")
        if self._connect_lock.locked():
            _LOGGER.warning("Connection is already in progress, skipping connect.")
            return
        if self.available:
            _LOGGER.warning("Device is connected, skipping connection.")
            return

        async with self._connect_lock:
            _LOGGER.debug("Opening Arcam connection.")
            if self._writer is not None:
                raise RuntimeError("Device already connected.")

            # open a connection
            reader, writer = await asyncio.wait_for(
                asyncio.open_connection(self._host, self._port),
                timeout=self._timeout
            )
            _LOGGER.info("Device connection established.")
            self._reader = reader
            self._writer = writer
            self.available = True
            self._reconnect = reconnect
            self._set_socket_options()

            await self._responder_cancel()
            await self._listener_schedule()
            await asyncio.sleep(0) # yield to listener task
            await self.discover_zones(self._enabled_zones) # discover zones
            await asyncio.sleep(2)
            await self._updater_schedule()

        _LOGGER.debug(">> ArcamSolo.connect() completed")

    async def disconnect(self):
        """Shutdown and close telnet connection to Arcam."""
        _LOGGER.debug(">> ArcamSolo.disconnect() started")

        if self._disconnect_lock.locked():
            _LOGGER.warning(
                "Arcam connection is already disconnecting, skipping disconnect"
            )
            return
        if not self.available:
            _LOGGER.warning("Arcam not connected, skipping disconnect")
            return

        async with self._disconnect_lock:
            _LOGGER.debug("disconnecting Arcam connection")
            self.available = False
            for zone in self.zones:
                self._call_zone_callbacks(zone)

            await self._listener_cancel()
            await self._responder_cancel()

            writer = self._writer
            if writer:
                # Close AVR connection
                _LOGGER.debug("closing Arcam connection")
                self._writer.close()
                try:
                    await self._writer.wait_closed()
                except Exception as exc:  # pylint: disable=broad-except
                    _LOGGER.debug("ignoring responder exception %s", str(exc))
            self._reader = None
            self._writer = None
            _LOGGER.info("Arcam connection closed")

            await self._reconnect_schedule()

        _LOGGER.debug(">> ArcamSolo.disconnect() completed")

    async def shutdown(self):
        """Shutdown the client."""
        _LOGGER.debug(">> ArcamSolo.shutdown()")
        self._reconnect = False
        await self._reconnect_cancel()
        await self.disconnect()

    async def reconnect(self):
        """Reconnect to an ArcamSolo."""
        _LOGGER.debug(">> ArcamSolo.reconnect() started")
        retry = 0
        try:
            while not self.available:
                delay = get_backoff_delay(retry)
                _LOGGER.debug("waiting %ds before retrying connection", delay)
                await asyncio.sleep(delay)

                retry += 1
                try:
                    await self.connect()
                    if self.available:
                        break
                except asyncio.CancelledError:  # pylint: disable=try-except-raise
                    # pass through to outer except
                    raise
                except Exception as exc:  # pylint: disable=broad-except
                    _LOGGER.debug(
                        "could not reconnect to Arcam Solo: %s: %s", type(exc).__name__, exc
                    )
                    # fall through to reconnect outside try block

                if self.available:
                    await self.disconnect()
        except asyncio.CancelledError:
            _LOGGER.debug(">> ArcamSolo.reconnect() cancelled")

        _LOGGER.debug(">> ArcamSolo.reconnect() completed")

    async def _reconnect_schedule(self):
        """Schedule reconnection to ArcamSolo."""
        if self._reconnect:
            _LOGGER.debug(">> ArcamSolo._reconnect_schedule()")
            reconnect_task = self._reconnect_task
            if reconnect_task:
                await asyncio.sleep(0)  # yield to reconnect task if running
                if reconnect_task.done():
                    reconnect_task = None  # trigger new task creation
            if reconnect_task is None:
                _LOGGER.info("reconnecting to ArcamSolo")
                reconnect_task = asyncio.create_task(self.reconnect())
                self._reconnect_task = reconnect_task
            else:
                _LOGGER.error("ArcamSolo listener reconnection already running")

    async def _reconnect_cancel(self):
        """Cancel any active reconnect task."""
        await cancel_task(self._reconnect_task, "reconnect")
        self._reconnect_task = None

    # Reader co-routine
    async def _reader_readuntil(self):
        """Read from reader with cancel detection."""
        try:
            return await self._reader.readuntil(ARCAM_COMM_END)
        except asyncio.CancelledError:
            _LOGGER.debug("reader: readuntil() was cancelled")
            return None

    async def _read_response(self, timeout=None):
        """Wait for a response from device and return to all readers."""
        _LOGGER.debug(">> ArcamSolo._read_response(timeout=%s)", timeout)

        # Schedule responder task if not already created
        responder_task = self._responder_task
        if responder_task:
            if responder_task.done():
                responder_task = None  # trigger new task creation
        if responder_task is None:
            responder_task = asyncio.create_task(self._reader_readuntil())
            self._responder_task = responder_task
            _LOGGER.debug(">> ArcamSolo._read_response() created responder task %s",
                          responder_task.get_name())
        else:
            # Wait on existing responder task
            _LOGGER.debug(">> ArcamSolo._read_response() using responder task %s",
                          responder_task.get_name())

        # Wait for result and process
        task_name = asyncio.current_task().get_name()
        try:
            if timeout:
                _LOGGER.debug(
                    ">> ArcamSolo._read_response() %s: waiting for data (timeout=%s)",
                    task_name,
                    timeout
                )
                done, pending = await asyncio.wait(  # pylint: disable=unused-variable
                    [responder_task], timeout=timeout
                )
                if done:
                    raw_response = responder_task.result()
                else:
                    _LOGGER.debug(">> ArcamSolo._read_response() %s: timed out waiting for data",
                                  task_name)
                    return None
            else:
                _LOGGER.debug(">> ArcamSolo._read_response() %s: waiting for data", task_name)
                raw_response = await responder_task
        except (EOFError, TimeoutError):
            # Connection closed
            _LOGGER.debug(">> ArcamSolo._read_response() %s: connection closed", task_name)
            return None
        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.error(">> ArcamSolo._read_response() %s: exception: %s", task_name, exc)
            return None
        if raw_response is None:  # task cancelled
            return None
        # Arcam works in bytes so we just pass the bytes around
        _LOGGER.debug(">> ArcamSolo._read_response() %s: received response: %s",
                      task_name,
                      raw_response)
        return raw_response

    async def discover_zones(self, zones: list[int]):
        """Discovers all available zones."""
        _LOGGER.debug('>> ArcamSolo.discover_zones(zones="%s")', zones)
        for z in zones:
            await self.send_raw_command(
                command="volume",
                data=[b'\xF0'],
                zone=z
            )
            await asyncio.sleep(0.1) # allow system to respond

    async def send_raw_command(self,
                               command: str,
                               data: list[bytes],
                               zone: int=1,
                               rate_limit=True):
        """Send a raw command to the Arcam."""
        _LOGGER.debug(
            '>> ArcamSolo.send_raw_command("%s", rate_limit=%s)',
            command,
            rate_limit,
        )
        if not self.available:
            raise RuntimeError("AVR connection not available")

        if rate_limit:
            # Rate limit commands
            command_delay = 0.1
            since_command = command_delay + 0.1
            if self._last_command:
                since_command = time.time() - self._last_command
            if since_command < command_delay:
                delay = command_delay - since_command
                _LOGGER.debug("delaying command for %.3f s", delay)
                await asyncio.sleep(command_delay - since_command)
        raw_command = COMMAND_CODES.get(command)
        raw_command = (
            ARCAM_COMM_START +
            zone.to_bytes(1, 'little') +
            raw_command +
            len(data).to_bytes(1, 'little')
        )
        for b in data:
            raw_command += b
        raw_command += ARCAM_COMM_END
        _LOGGER.debug("sending command: %s", raw_command)
        self._writer.write(raw_command)
        await self._writer.drain()
        self._last_command = time.time()

    async def send_ir_command(self,
                              command: str):
        """Send an IR command to the Arcam."""
        ir_data = IR_COMMAND_CODES.get(command, None)
        if ir_data is None:
            raise ValueError("Command does not exist.")
        await self.send_raw_command(
            command="virtual_remote",
            data=[
                ir_data["system_code"].to_bytes(1, 'little'),
                ir_data["command_code"].to_bytes(1, 'little')
            ]
        )

    async def set_source(self,
                         source: str):
        """Set the source of the system."""
        await self.send_ir_command(
            SOURCE_IR_CONTROL_MAP.get(source)
        )

    async def set_volume(self,
                         volume: int):
        """Set the volume level of the system."""
        if volume > 72:
            raise ValueError("Max volume is 72.")
        await self.send_raw_command(
            command="volume",
            data=[
                volume.to_bytes(1, 'little')
            ]
        )

    async def mute_on(self):
        """Turn mute on."""
        await self.send_ir_command(
            command="mute_on"
        )

    async def mute_off(self):
        """Turn mute on."""
        await self.send_ir_command(
            command="mute_off"
        )

    async def turn_off(self):
        """Turn the unit off."""
        await self.send_ir_command(
            command="standby_on"
        )

    async def turn_on(self):
        """Turn the unit on."""
        # send the command twice as sometimes the device doesn't respond
        await self.send_ir_command(
            command="standby_off"
        )
        await self.send_ir_command(
            command="standby_off"
        )

    @property
    def source(self) -> str | None:
        """Return the current input source."""
        if 1 not in self.zones:
            return None
        return self.zones.get(1).get("source", None)



"""Arcam Parser Helper."""

import logging

from .commands import (
    SOURCE_SELECTION_CODES,
    ANSWER_CODES,
    ACCEPTED_ANSWER_CODES,
    COMMAND_CODES,
    POWER_STATUS_CODES,
    CD_PLAYBACK_STATUS_CODES,
    RADIO_QUERY_COMMANDS,
    RADIO_MPEG_MODES
)

_LOGGER = logging.getLogger(__name__)
CURRENT_SOURCE = ""

def get_answer_code(ac: bytes) -> str:
    """Return the answer code from the byte."""
    return (list(ANSWER_CODES.keys())[list(ANSWER_CODES.values()).index(ac)])

def get_command_code(cc: bytes) -> str:
    """Return the command code from the byte."""
    return (list(COMMAND_CODES.keys())[list(COMMAND_CODES.values()).index(cc)])

def get_radio_query_code(d1: bytes) -> str:
    """Return the command name from the byte."""
    return (list(RADIO_QUERY_COMMANDS.keys())[list(RADIO_QUERY_COMMANDS.values()).index(d1)])

def parse_response(response: bytes) -> dict | list[dict] | None:
    """Convert response bytes into a tuple for the main module to handle."""
    global CURRENT_SOURCE
    output = {
        "k": "",
        "v": None,
        "z": None
    }
    # Ignore start byte
    output["z"] = response[1:2][0] # Second byte is zone
    cc = response[2:3] # Third byte is command code
    ac = response[3:4] # Forth byte is answer code
    size = response[4:5][0] # Fith byte is length
    data = response[5:(5+size)] # Sixth byte+ is data

    cc = get_command_code(cc)
    ac = get_answer_code(ac)
    # check answer code is valid
    if ac not in ACCEPTED_ANSWER_CODES:
        if ac == "command_invalid_at_this_time":
            _LOGGER.warning("Command %s is unavailable in current state. Will retry later.", cc)
            return None
        raise ValueError(
            f"Provided response for {cc} is invalid at this time: {get_answer_code(ac)}"
        )

    if cc == "volume":
        output["k"] = cc
        output["v"] = data[0]
    elif cc == "mute":
        output["k"] = "muted"
        output["v"] = bool(data[0])
    elif cc == "source":
        output["k"] = "source"
        output["v"] = SOURCE_SELECTION_CODES.get(data)
        CURRENT_SOURCE = output["v"]
    elif cc == "status":
        output["k"] = "power"
        output["v"] = POWER_STATUS_CODES.get(data)
    elif cc == "software_version":
        output["k"] = cc
        output["v"] = f"{data[0]}.{data[1]}"
    elif cc == "rs232_version":
        output["k"] = cc
        output["v"] = f"{data[0]}.{data[1]}"
    elif cc == "balance":
        output["k"] = cc
        output["v"] = (data[0])-100
    elif cc == "bass":
        output["k"] = cc
        output["v"] = bytes_to_int_with_offset(data[0], 2, 0x5D,
                                               range_upper_limit=14)
    elif cc == "treble":
        output["k"] = cc
        output["v"] = bytes_to_int_with_offset(data[0], 2, 0x5D,
                                               range_upper_limit=14)
    elif cc == "display_brightness":
        output["k"] = cc
        output["v"] = data[0]
    elif cc == "stby_display_brightness":
        output["k"] = "standby_display_brightness"
        output["v"] = data[0]
    elif cc == "cd_playback_state":
        output["k"] = cc
        output["v"] = CD_PLAYBACK_STATUS_CODES.get(data, "Unknown")
    elif cc == "cdusb_playback_time":
        output["k"] = "current_track_position"
        hour = int.from_bytes(data[0:1]) * 3600
        minute = int.from_bytes(data[1:2]) * 60
        sec = int.from_bytes(data[2:3])
        output["v"] = hour+minute+sec
    elif cc == "cdusb_current_track":
        return parse_cdusb_current_track(
            z=output["z"],
            b=data
        )
    elif cc == "radio_station_info":
        return parse_radio_station_info(
            z=output["z"],
            b=data
        )
    elif cc == "radio_station":
        output["k"] = "radio_station"
        output["v"] = data.decode("ascii")

    return output if output["v"] is not None else None

def parse_radio_station_info(z: int, b: bytes) -> list[dict]:
    """Parse radio ration information."""
    global CURRENT_SOURCE
    d1 = get_radio_query_code(b[0:1])
    if d1 == "request_station_frequency":
        mhz = int.from_bytes(b[1:2])
        khz = int.from_bytes(b[2:3])
        if CURRENT_SOURCE == "AM":
            val = str(mhz) + str(khz)
        elif CURRENT_SOURCE == "FM":
            val = int(str(mhz) + str(khz).zfill(2))/100
        else:
            val = None
        return {
            "k": "radio_frequency",
            "v": val,
            "z": z
        }
    elif d1 == "request_station_signal":
        return {
            "k": "radio_signal",
            "v": int.from_bytes(b[1:2]),
            "z": z
        }
    elif d1 == "request_mpeg_mode":
        return {
            "k": "dab_mpeg_mode",
            "v": RADIO_MPEG_MODES.get(b[1:2], None),
            "z": z
        }
    elif d1 == "request_data_rate":
        return {
            "k": "dab_data_rate",
            "v": int.from_bytes(b[1:2]),
            "z": z
        }
    return None

def parse_cdusb_current_track(z: int, b: bytes) -> list[dict]:
    """Parse CD / USB Current Track info."""
    current_folder = int.from_bytes(b[0:1])
    total_folder = int.from_bytes(b[1:2])
    msb_current_track = int.from_bytes(b[2:3])
    lsb_current_track = int.from_bytes(b[3:4])
    msb_total_track = int.from_bytes(b[4:5])
    lsb_total_track = int.from_bytes(b[5:6])
    return [
        {
            "k": "current_folder",
            "v": current_folder,
            "z": z
        },
        {
            "k": "total_folder",
            "v": total_folder,
            "z": z
        },
        {
            "k": "msb_current_track",
            "v": msb_current_track,
            "z": z
        },
        {
            "k": "lsb_current_track",
            "v": lsb_current_track,
            "z": z
        },
        {
            "k": "msb_total_track",
            "v": msb_total_track,
            "z": z
        },
        {
            "k": "lsb_total_track",
            "v": lsb_total_track,
            "z": z
        },
    ]

def bytes_to_int_with_offset(b: bytes, offset: int, normalizer, range_upper_limit: int):
    """Converts bytes to an int with a provided offset"""
    i = int.from_bytes(b, byteorder='big', signed=True)
    # apply a base range from a normalizer
    i -= normalizer
    # now apply scaling using the offset provided
    i *= offset
    i -= range_upper_limit
    return i



"""ArcamSolo utils."""

import asyncio
import random
import socket
import logging

RECONNECT_DELAY_MAX = 64

_LOGGER = logging.getLogger(__name__)

## Source: https://stackoverflow.com/questions/12248132/how-to-change-tcp-keepalive-timer-using-python-script  # pylint: disable=line-too-long
def sock_set_keepalive(sock, after_idle_sec=1, interval_sec=3, max_fails=5):
    """Set TCP keepalive on an open socket.

    It activates after 1 second (after_idle_sec) of idleness,
    then sends a keepalive ping once every 3 seconds (interval_sec),
    and closes the connection after 5 failed ping (max_fails), or 15 seconds
    """
    if sock:
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)
        except:  # pylint: disable=bare-except
            pass  ## ignore if keepalive cannot be set

async def safe_wait_for(awt, timeout=None):
    """
    asyncio.wait_for() that re-raises cancellation even if aw is complete.
    Work around issue: https://bugs.python.org/issue42130
    """
    task = asyncio.create_task(awt)
    try:
        await asyncio.wait({task}, timeout=timeout)
        if task.done():
            return task.result()
        ## timed out
        try:
            task.cancel()
            await task
        except asyncio.CancelledError:
            ## TODO: what if wait_for is cancelled when waiting for task to be cancelled?
            pass
        raise asyncio.TimeoutError()
    except asyncio.CancelledError as exc:
        raise exc

def get_backoff_delay(retry_count):
    """Calculate exponential backoff with random jitter delay."""
    delay = round(
        min(RECONNECT_DELAY_MAX, (2 ** max(retry_count, 2)))
        + (random.randint(0, 1000) / 1000),
        4,
    )
    return delay

async def cancel_task(task, task_name=None, debug=False):
    """Cancels a task and waits for it to finish."""
    if task:
        if task_name is None:
            task_name = task.get_name()
        ## Trap exception if not called from a task
        current_task = None
        try:
            current_task = asyncio.Task.current_task()
        except AttributeError:
            pass
        if current_task is not None and task == current_task:
            if debug:
                _LOGGER.debug(
                    ">> cancel_task(%s): ignoring cancellation of current task",
                    task_name,
                )
            return
        if not task.done():
            if debug:
                _LOGGER.debug(">> cancel_task(%s): requesting cancellation", task_name)
            task.cancel()
            if debug:
                _LOGGER.debug(">> cancel_task(%s): waiting for completion", task_name)
            try:
                await task
            except asyncio.CancelledError:
                if debug:
                    _LOGGER.debug(">> cancel_task(%s): cancelled exception", task_name)
            except Exception as exc:  # pylint: disable=broad-except
                if debug:
                    _LOGGER.debug(
                        ">> cancel_task(%s): returned exception: %s", task_name, exc
                    )
            else:
                if debug:
                    _LOGGER.debug(">> cancel_task(%s): completed", task_name)
        else:
            if debug:
                _LOGGER.debug(">> cancel_task(%s): already completed", task_name)
            exc = task.exception()
            if exc:
                if debug:
                    _LOGGER.debug(
                        ">> cancel_task(%s): returned exception: %s", task_name, exc
                    )



"""Arcam Solo commands."""

ARCAM_COMM_START = b'\x21'
ARCAM_COMM_END = b'\x0D'

COMMAND_CODES = {
    "status": b'\x00',
    "source": b'\x01',
    "volume": b'\x02',
    "balance": b'\x04',
    "bass": b'\x05',
    "treble": b'\x06',
    "mute": b'\x08',
    "preset_select": b'\x13',
    "display_brightness": b'\x0A',
    "stby_display_brightness": b'\x16',
    "snooze": b'\x29',
    "sleep": b'\x31',
    "headphones": b'\x32',
    "tuner_type": b'\x37',
    "radio_station": b'\xDE',
    "radio_station_info": b'\xDF',
    "radio_programme_type": b'\xE6',
    "radio_rds_information": b'\xE7',
    "virtual_remote": b'\xE9',
    "cd_playback_state": b'\xEC',
    "cd_play_mode": b'\x54',
    "usb_playback_state": b'\x58',
    "usb_play_mode": b'\x59',
    "cdusb_total_track_time": b'\x55',
    "cdusb_total_playback_time": b'\x56',
    "preset_save": b'\xDC',
    "cdusb_current_track": b'\xE8',
    "usb_track_name": b'\x60',
    "usb_folder_name": b'\x61',
    "cdusb_playback_time": b'\xEB',
    "cd_disc_type": b'\xED',
    "time": b'\xF8',
    "rs232_version": b'\xF1',
    "software_version": b'\xF2',
}

ARCAM_QUERY_COMMANDS = list(COMMAND_CODES.keys())

RADIO_QUERY_COMMANDS = {
    "request_station_frequency": b'\x00', # fm/am
    "request_station_signal": b'\x01', # all
    "request_mpeg_mode": b'\x02', # dab only
    "request_data_rate": b'\x03' # dab only
}

ACCEPTED_ANSWER_CODES = ["status_update", "command_accepted_completed"]
ANSWER_CODES = {
    "status_update": b'\x00',
    "command_accepted_completed": b'\x01',
    "command_accepted_processing": b'\x02',
    "zone_invalid": b'\x82',
    "command_not_recognised": b'\x83',
    "parameter_not_recognised": b'\x84',
    "command_invalid_at_this_time": b'\x85',
    "invalid_data_length": b'\x86'
}

SOURCE_SELECTION_CODES = {
    b'\x00': "N/A",
    b'\x01': "FM",
    b'\x02': "DAB",
    b'\x03': "TAPE",
    b'\x04': "AV",
    b'\x05': "N/A",
    b'\x06': "N/A",
    b'\x07': "AM",
    b'\x08': "GAME",
    b'\x09': "USB",
    b'\x0A': "CD",
    b'\x0B': "TV",
    b'\x0C': "AUX"
}

SOURCE_IR_CONTROL_MAP = {
    "FM": "src_fm",
    "AM": "src_am",
    "DAB": "src_dab",
    "TAPE": "src_tape",
    "AV": "src_av",
    "GAME": "src_game",
    "USB": "src_usb",
    "CD": "src_cd",
    "TV": "src_tv",
    "AUX": "src_aux"
}

POWER_STATUS_CODES = {
    b'\x00': "Standby",
    b'\x01': "Power on",
    b'\x02': "Alarm",
    b'\x03': "Basic Menu",
    b'\x04': "N/A",
    b'\x05': "Clock Menu",
    b'\x06': "initialising"
}

DISPLAY_BRIGHTNESS_CODES = {
    b'\x00': "Off",
    b'\x01': "25%",
    b'\x02': "50%",
    b'\x03': "75%",
    b'\x04': "100%"
}

TUNER_MODULE_TYPE_CODES = {
    b'\x00': "UK DAB/FM/AM",
    b'\x01': "US AM/FM",
    b'\x02': "Japan",
    b'\x03': "EU"
}

CD_PLAYBACK_STATUS_CODES = {
    b'\x01': "Loading",
    b'\x02': "Playing",
    b'\x03': "Stopped",
    b'\x04': "Scanning Back",
    b'\x05': "Scanning Forward",
    b'\x08': "Tray Open / Empty",
    b'\x09': "Paused",
    b'\x0D': "Track Skipping"
}

USB_PLAYBACK_STATUS_CODES = {
    b'\x01': "Initialising",
    b'\x02': "Playing",
    b'\x03': "Stopped",
    b'\x04': "Scanning Back",
    b'\x05': "Scanning Forward",
    b'\x08': "No Device",
    b'\x09': "Paused",
    b'\x10': "Invalid File",
    b'\x11': "No Valid Files Present",
    b'\x0D': "Track Skipping"
}

RADIO_MPEG_MODES = {
    b'\x00': "Stereo",
    b'\x01': "Joint Stereo",
    b'\x02': "Dual Mono",
    b'\x03': "Mono"
}

IR_COMMAND_CODES = {
    "src_tv": {
        "system_code": 16,
        "command_code": 0
    },
    "src_av": {
        "system_code": 16,
        "command_code": 2
    },
    "src_dab": {
        "system_code": 16,
        "command_code": 3
    },
    "src_tape": {
        "system_code": 16,
        "command_code": 5
    },
    "src_cd": {
        "system_code": 16,
        "command_code": 7
    },
    "src_aux": {
        "system_code": 16,
        "command_code": 8
    },
    "standby": {
        "system_code": 16,
        "command_code": 12
    },
    "mute": {
        "system_code": 16,
        "command_code": 13
    },
    "volume_plus": {
        "system_code": 16,
        "command_code": 16
    },
    "volume_minus": {
        "system_code": 16,
        "command_code": 17
    },
    "bass_plus": {
        "system_code": 16,
        "command_code": 22
    },
    "bass_minus": {
        "system_code": 16,
        "command_code": 23
    },
    "treble_plus": {
        "system_code": 16,
        "command_code": 24
    },
    "treble_minus": {
        "system_code": 16,
        "command_code": 25
    },
    "balance_right": {
        "system_code": 16,
        "command_code": 26
    },
    "balance_left": {
        "system_code": 16,
        "command_code": 27
    },
    "src_am": {
        "system_code": 16,
        "command_code": 52
    },
    "src_fm": {
        "system_code": 16,
        "command_code": 53
    },
    "display_brightness": {
        "system_code": 16,
        "command_code": 59
    },
    "navigate_right": {
        "system_code": 16,
        "command_code": 80
    },
    "navigate_left": {
        "system_code": 16,
        "command_code": 81
    },
    "menu": {
        "system_code": 16,
        "command_code": 82
    },
    "navigate_down": {
        "system_code": 16,
        "command_code": 85
    },
    "navigate_up": {
        "system_code": 16,
        "command_code": 86
    },
    "ok": {
        "system_code": 16,
        "command_code": 87
    },
    "alarm_1_toggle": {
        "system_code": 16,
        "command_code": 113
    },
    "alarm_2_toggle": {
        "system_code": 16,
        "command_code": 114
    },
    "alarm_3_toggle": {
        "system_code": 16,
        "command_code": 115
    },
    "alarm_4_toggle": {
        "system_code": 16,
        "command_code": 116
    },
    "snooze": {
        "system_code": 16,
        "command_code": 117
    },
    "sleep": {
        "system_code": 16,
        "command_code": 118
    },
    "mute_on": {
        "system_code": 16,
        "command_code": 119
    },
    "mute_off": {
        "system_code": 16,
        "command_code": 120
    },
    "standby_off": {
        "system_code": 16,
        "command_code": 123
    },
    "standby_on": {
        "system_code": 16,
        "command_code": 124
    },
    "cd_play": {
        "system_code": 20,
        "command_code": 53
    },
    "cd_pause": {
        "system_code": 20,
        "command_code": 48
    },
    "cd_stop": {
        "system_code": 20,
        "command_code": 54
    },
    "cd_eject": {
        "system_code": 20,
        "command_code": 45
    },
    "cd_track_next": {
        "system_code": 20,
        "command_code": 32
    },
    "cd_track_previous": {
        "system_code": 20,
        "command_code": 33
    },
}
